#!/usr/bin/env python
"""
ELF Parser
by Javantea
Aug 2007 for virus8a.py

Simply read an ELF file. No funny business.

Currently only reads 32-bit and 64-bit ELF files in little endian.

32-bit support is not tested yet but should be in about 10 minutes.

"""
from __future__ import print_function
import struct
import sys
python3 = (sys.hexversion >= 0x3000000)

# From /usr/include/elf.h:
#   Copyright (C) 1995-2014 Free Software Foundation, Inc.
# GNU Lesser General Public
#/* Legal values for ST_TYPE subfield of st_info (symbol type).  */

STT_NOTYPE = 	0		#/* Symbol type is unspecified */
STT_OBJECT = 	1		#/* Symbol is a data object */
STT_FUNC = 	2		#/* Symbol is a code object */
STT_SECTION = 	3		#/* Symbol associated with a section */
STT_FILE = 	4		#/* Symbol's name is file name */
STT_COMMON = 	5		#/* Symbol is a common data object */
STT_TLS = 	6		#/* Symbol is thread-local data object*/
STT_NUM =	7		#/* Number of defined types.  */
STT_LOOS = 	10		#/* Start of OS-specific */
STT_GNU_IFUNC = 10		#/* Symbol is indirect code object */
STT_HIOS = 	12		#/* End of OS-specific */
STT_LOPROC = 	13		#/* Start of processor-specific */
STT_HIPROC = 	15		#/* End of processor-specific */

ELFCLASS32 = 1
ELFCLASS64 = 2

class elf_header:
	"""
	/sbin/grub 32-bit x86
00000000  7f 45 4c 46 01 01 01 00  00 00 00 00 00 00 00 00  |.ELF............|
	simple     64-bit x86-64
00000000  7f 45 4c 46 02 01 01 00  00 00 00 00 00 00 00 00  |.ELF............|
	
	"""
	def __init__(self, data, error=None):
		# This is bad news. We could encode, but I prefer them acting like sane programmers.
		if python3 and type(data) == str:
			if error == None:
				raise ValueError("ELF header requires bytes, not a string.")
			#end if
			# default values please.
			self.default()
			return
		#end if
		if len(data) < 52:
			if error == None:
				raise ValueError("Data is not long enough, " + 
				"we need at least 52 bytes.")
			#end if
			self.default()
			return
		#end if
		# Class tells us whether it's 64-bit or 32-bit.
		self.elfclass = data[4]
		if not python3: self.elfclass = ord(self.elfclass)
		# parse data, version, osabi, abiversion, pad
		endian = data[5]
		if not python3: endian = ord(endian)
		version = data[6]
		if not python3: version = ord(version)
		osabi = data[7]
		if not python3: osabi = ord(osabi)
		abiversion = data[8]
		if not python3: abiversion = ord(abiversion)
		pad = data[9:16]
		# 64-bit
		if self.elfclass == ELFCLASS64:
			if len(data) < 64:
				if error == None:
					raise ValueError("Data is not long enough, " + 
						"we need at least 64 bytes.")
				#end if
				self.default()
				return
			#end if
			len1 = 64
			(self.e_ident, self.e_type, self.e_machine, self.e_version, self.e_entry, 
				self.e_phoff, self.e_shoff, self.e_flags, self.e_ehsize, 
				self.e_phentsize, self.e_phnum, self.e_shentsize, self.e_shnum, 
				self.e_shstrndx) = struct.unpack('<16sHHLQQQLHHHHHH', data[:len1])
		elif self.elfclass == ELFCLASS32:
			len1 = 52
			# 32-bit
			(self.e_ident, self.e_type, self.e_machine, self.e_version, self.e_entry, 
				self.e_phoff, self.e_shoff, self.e_flags, self.e_ehsize, 
				self.e_phentsize, self.e_phnum, self.e_shentsize, self.e_shnum, 
				self.e_shstrndx) = struct.unpack('<16sHHLLLLLHHHHHH', data[:len1])
		else:
			if error == None:
				raise ValueError("ELF Class is invalid.")
			#end if
			# Guess that it's 64-bit unless they say error='32'
			if error[:2] == '32':
				self.elfclass = 1
				len1 = 52
				# 32-bit
				(self.e_ident, self.e_type, self.e_machine, self.e_version, self.e_entry, 
					self.e_phoff, self.e_shoff, self.e_flags, self.e_ehsize, 
					self.e_phentsize, self.e_phnum, self.e_shentsize, self.e_shnum, 
					self.e_shstrndx) = struct.unpack('<16sHHLLLLLHHHHHH', data[:len1])
			else:
				self.elfclass = 2
				len1 = 64
				(self.e_ident, self.e_type, self.e_machine, self.e_version, self.e_entry, 
					self.e_phoff, self.e_shoff, self.e_flags, self.e_ehsize, 
					self.e_phentsize, self.e_phnum, self.e_shentsize, self.e_shnum, 
					self.e_shstrndx) = struct.unpack('<16sHHLQQQLHHHHHH', data[:len1])
			#end if
			return
		#end if
	#end def __init__(data)
	
	def default(self):
		self.elfclass = 2
		
		# A value from /bin/ls
		(self.e_ident, self.e_type, self.e_machine, self.e_version, 
			self.e_entry, self.e_phoff, self.e_shoff, self.e_flags, 
			self.e_ehsize, self.e_phentsize, self.e_phnum, 
			self.e_shentsize, self.e_shnum, self.e_shstrndx) = \
			(b'\x7fELF\x02\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00', 
			2, 62, 1, 4212636, 64, 112336, 0, 64, 56, 10, 64, 27, 
			26)
	#end def default()
	
	def out(self):
		data1 = 'LLL'
		if self.elfclass == ELFCLASS64:
			data1 = 'QQQ'
		#end if
		return self.e_ident + struct.pack('<HHL' + data1 + 'LHHHHHH', self.e_type, 
			self.e_machine, self.e_version, 
			self.e_entry, self.e_phoff, self.e_shoff, self.e_flags, 
			self.e_ehsize, self.e_phentsize, self.e_phnum, self.e_shentsize, 
			self.e_shnum, self.e_shstrndx)
	#end def __init__(data)
	
	def __str__(self):
		return str((self.e_ident, self.e_type, self.e_machine, self.e_version, 
			self.e_entry, self.e_phoff, self.e_shoff, self.e_flags, 
			self.e_ehsize, self.e_phentsize, self.e_phnum, self.e_shentsize, 
			self.e_shnum, self.e_shstrndx))
	#end def __str__()
#end class elf_header


class section_header:
	def __init__(self, data, elf, i, error=None):
		self.elfclass = elf.elfclass
		if self.elfclass == ELFCLASS32:
			data1 = 'LLLLLLLLLL'
			len1 = 40
		elif self.elfclass == ELFCLASS64:
			data1 = 'LLQQQQLLQQ'
			len1 = 64
		else:
			raise ValueError('Not sure what type of ELF we\'re working on.')
		#end if
		# This could be done better.
		pos = elf.e_shoff + i * len1
		if len(data) < pos+len1:
			if error == None:
				raise ValueError("We don't have enough data to load " + 
					"the section header: %08x." % (pos+len1))
			#end if
			if self.elfclass == ELFCLASS64:
				(self.sh_name, self.sh_type, self.sh_flags, self.sh_addr, 
					self.sh_offset, self.sh_size, self.sh_link, self.sh_info, 
					self.sh_addralign, 
					self.sh_entsize) = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
			else:
				(self.sh_name, self.sh_type, self.sh_flags, self.sh_addr, 
					self.sh_offset, self.sh_size, self.sh_link, self.sh_info, 
					self.sh_addralign, 
					self.sh_entsize) = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
			#end if
			self.sh_name_str = ''
			return
		#end if
		(self.sh_name, self.sh_type, self.sh_flags, self.sh_addr, 
			self.sh_offset, self.sh_size, self.sh_link, self.sh_info, 
			self.sh_addralign, 
			self.sh_entsize) = struct.unpack('<' + data1, data[pos:pos+len1])
		
		self.sh_name_str = ''
	#end def __init__(data, elf)
	
	def setName(self, data, sect):
		"""
		Dereference the data.
		"""
		# FIXME: 320 assumes there won't be a string longer than 320 characters.
		self.sh_name_str = data[sect.sh_offset+self.sh_name:sect.sh_offset+self.sh_name+320]
		self.sh_name_str = self.sh_name_str[:self.sh_name_str.index(b'\x00')]
	#end def setName(data, sect)
	
	def setFlags(self, w, a, x):
		"""
		Write, alloc, execute
	sh_flags  Sections  support one-bit flags that describe miscellaneous attributes.  If a flag
		bit is set in sh_flags, the attribute is "on" for  the  section.   Otherwise,  the
		attribute is "off" or does not apply.  Undefined attributes are set to zero.
		"""
		# See man 5 elf sh_flags SHF_*.
		self.sh_flags = w * 1 + a * 2 + x * 4
	#end def setFlags(w, a, x)
	
	def out(self):
		data1 = 'LLLLLLLLLL'
		if self.elfclass == ELFCLASS64:
			data1 = 'LLQQQQLLQQ'
		#end if
		return struct.pack('<' + data1, self.sh_name, self.sh_type, 
			self.sh_flags, self.sh_addr, 
			self.sh_offset, self.sh_size, self.sh_link, self.sh_info, 
			self.sh_addralign, self.sh_entsize)
	#end def out()
	
	def __str__(self):
		return str((self.sh_name, self.sh_type, self.sh_flags, self.sh_addr, 
			self.sh_offset, self.sh_size, self.sh_link, self.sh_info, 
			self.sh_addralign, self.sh_entsize, self.sh_name_str))
	#end def __str__()
	
#end class section_header

class prog_header:
	def __init__(self, data, elf, i, error=None):
		self.elfclass = elf.elfclass
		if elf.elfclass == ELFCLASS32:
			len1 = 32
		elif elf.elfclass == ELFCLASS64:
			len1 = 56
		else:
			raise ValueError('Not sure what type of ELF we\'re working on.')
		#end if
		
		# This doesn't seem right. Who wrote this? Oh yeah, I did.
		pos = elf.e_phoff + i * len1
		if len(data) < pos+len1:
			if error == None:
				raise ValueError("We don't have enough data to load " + 
					"the program header: %08x." % (pos+len1))
			#end if
			if elf.elfclass == ELFCLASS64:
				# Default values for 64-bit.
				(self.p_type, self.p_flags, self.p_offset, self.p_vaddr, 
					self.p_paddr, self.p_filesz, self.p_memsz, 
					self.p_align) = (6, 5, 64, 0x400040, 
					0x400040, 560, 560, 8)
			else:
				# Default values for 32-bit.
				(self.p_type, self.p_offset, self.p_vaddr, self.p_paddr, 
					self.p_filesz, self.p_memsz, self.p_flags, 
					self.p_align) = (6, 52, 0x8048034, 
					0x8048034, 320, 320, 5, 4)
			#end if
			return
		#end if
		if elf.elfclass == ELFCLASS64:
			# different sequence, silliness.
			(self.p_type, self.p_flags, self.p_offset, self.p_vaddr, 
				self.p_paddr, self.p_filesz, self.p_memsz, 
				self.p_align) = struct.unpack('<LLQQQQQQ', data[pos:pos+len1])
		else:
			(self.p_type, self.p_offset, self.p_vaddr, self.p_paddr, 
				self.p_filesz, self.p_memsz, self.p_flags, 
				self.p_align) = struct.unpack('<LLLLLLLL', data[pos:pos+len1])
		#end if
	#end def __init__(data, elf)
	
	def out(self):
		if self.elfclass == ELFCLASS64:
			return struct.pack('<LLQQQQQQ', self.p_type, self.p_flags, 
				self.p_offset, self.p_vaddr, self.p_paddr, self.p_filesz, 
				self.p_memsz, self.p_align)
		#end if
		return struct.pack('<LLLLLLLL', self.p_type, self.p_offset, 
			self.p_vaddr, self.p_paddr, self.p_filesz, self.p_memsz, 
			self.p_flags, self.p_align)
	#end def out()
	
	def __str__(self):
		return str((self.p_type, self.p_offset, self.p_vaddr, self.p_paddr, 
			self.p_filesz, self.p_memsz, self.p_flags, self.p_align))
	#end def __str__()
	
#end class prog_header

def parseELF(data, error=None):
	elf = elf_header(data, error=error)
	#print elf
	
	if elf.elfclass == ELFCLASS64:
		# I found out that the base address of elf64 files are 0x400000
		base_addr = 0x400000
	#end if
	
	#print 'Section Headers:'
	sections = []
	#for (i = 0, section = section_headers;
	#       i < elf_header.e_shnum;
	#       i++, section++)
	for i in range(elf.e_shnum):
		sections.append(section_header(data, elf, i))
	#next i
	for s in sections:
		s.setName(data, sections[elf.e_shstrndx])
	#next s
	#for s in sections: print str(s)
	
	#print 'Program Headers:'
	progs = []
	for i in range(elf.e_phnum):
		progs.append(prog_header(data, elf, i))
	#next i
	#for p in progs: print str(p)

	return (elf, sections, progs)
#end def parseElf(data)

def main():
	import sys
	filename = '/bin/ls'
	if len(sys.argv) > 1:
		filename = sys.argv[1]
	#end if
	elf, sections, progs = parseELF(open(filename, 'rb').read())
	print("ELF header (%s):" % (elf.elfclass == ELFCLASS32 and '32-bit' or '64-bit'))
	print(elf)
	print("Section Headers:")
	print('\n'.join([str(sec) for sec in sections]))
	print("Program Headers:")
	print('\n'.join([str(prog) for prog in progs]))
#end def main()

if __name__ == '__main__':
	main()
#end if
